/**
 * Copyright (C) 2007-2014 Linutop SAS <http://www.linutop.com/>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <gmodule.h>
#include <glib.h>
#include <glib/gprintf.h>
#include <libxfce4util/libxfce4util.h>

#include "ltop-heading.h"
#include "as-utils.h"
#include "as-chooser.h"

#define MARGIN 2
#define DESKTOPDIR DATADIR "/appstarter/plugins"

enum {
    STORE_COMPT = 0,
    STORE_ICON,
    STORE_NAME,
    STORE_ITEM,
};

typedef GtkWidget * (*GetWidgetFunc) (const gchar *file);
typedef gchar     * (*GetCmdFunc)    (void);
typedef void        (*FreeFunc)      (void);

typedef struct _App
{
    gchar *name;
    gchar *icon;
    gchar *display_name;
    GModule *module;
} app;

/*
 * prototypes
 */
static GList *build_modules_list ();
static app   *get_active_plugin (ASChooser *chooser);
static void   populate_store     (ASChooser *chooser);
static void   append_to_store    (gpointer data, gpointer user_data);

static void read_config (ASChooser *chooser);

static void plugin_changed   (GtkWidget *combo, gpointer data);
static void test_btn_clicked (GtkWidget *w, gpointer data);

static gboolean free_store (GtkTreeModel *model, GtkTreePath *path,
                            GtkTreeIter *iter,gpointer data);

/*
 * gobject fun
 */
struct _ASChooserPrivate
{
    GtkWidget *combo;
    GtkWidget *test_btn;
    GtkWidget *config;
//    GtkWidget *onlyme;
    GtkWidget *killss;
    GtkWidget *hidecursor;
    GtkWidget *delay;
    LtopHeading  *head;

    GtkListStore *store;
    GList        *modules;

    app         *active_item;
    const gchar *selected;
    const gchar *conffile;
};

static void as_chooser_class_init (ASChooserClass *klass);
static void as_chooser_init (ASChooser *chooser);
static void as_chooser_finalize (GObject *object);

G_DEFINE_TYPE (ASChooser, as_chooser, GTK_TYPE_VBOX);

static void
as_chooser_class_init (ASChooserClass *klass)
{
    g_type_class_add_private (klass, sizeof (ASChooserPrivate));

    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
    gobject_class->finalize = as_chooser_finalize;
}

static void
as_chooser_init (ASChooser *chooser)
{
    GtkWidget *hbox, *label, *hs;
    GtkCellRenderer *renderer;

    chooser->priv = G_TYPE_INSTANCE_GET_PRIVATE (chooser,
                                                 AS_TYPE_CHOOSER,
                                                 ASChooserPrivate);

    chooser->priv->active_item = NULL;

    /* interface */
    hbox = gtk_box_new (GTK_ORIENTATION_HORIZONTAL, MARGIN);
    gtk_box_pack_start (GTK_BOX (chooser), hbox, FALSE, FALSE, MARGIN);
    gtk_widget_show (hbox);

    label = gtk_label_new (_("Select an item: "));
    gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, MARGIN);
    gtk_widget_show (label);

    chooser->priv->store = gtk_list_store_new (4, G_TYPE_INT,
                                                  GDK_TYPE_PIXBUF,
                                                  G_TYPE_STRING,
                                                  G_TYPE_POINTER);
    chooser->priv->combo = gtk_combo_box_new_with_model (GTK_TREE_MODEL (
                                                         chooser->priv->store));
    renderer = gtk_cell_renderer_pixbuf_new ();
    g_object_set (renderer, "width", 28, NULL);
    gtk_cell_layout_pack_start (GTK_CELL_LAYOUT (chooser->priv->combo),
                                renderer, FALSE);
    gtk_cell_layout_add_attribute (GTK_CELL_LAYOUT (chooser->priv->combo),
                                   renderer, "pixbuf", STORE_ICON);

    renderer = gtk_cell_renderer_text_new ();
    gtk_cell_layout_pack_start (GTK_CELL_LAYOUT (chooser->priv->combo),
                                renderer, TRUE);
    gtk_cell_layout_add_attribute (GTK_CELL_LAYOUT (chooser->priv->combo),
                                   renderer, "text", STORE_NAME);

    gtk_box_pack_start (GTK_BOX (hbox), chooser->priv->combo,
                        FALSE, FALSE, MARGIN);
    gtk_widget_show (chooser->priv->combo);

    chooser->priv->test_btn = gtk_button_new_with_label (_("Run a test"));
    gtk_button_set_image (GTK_BUTTON (chooser->priv->test_btn),
                          gtk_image_new_from_stock (GTK_STOCK_MEDIA_PLAY,
                                                    GTK_ICON_SIZE_BUTTON));
    gtk_box_pack_end (GTK_BOX (hbox), chooser->priv->test_btn,
                      FALSE, FALSE, MARGIN);
    gtk_widget_show (chooser->priv->test_btn);

/*
    chooser->priv->onlyme = gtk_check_button_new_with_label (
        _("Start only this application (not the full desktop)"));
    gtk_box_pack_start (GTK_BOX (chooser), chooser->priv->onlyme,
                        FALSE, FALSE, MARGIN);
    gtk_widget_show (chooser->priv->onlyme);
*/

    chooser->priv->killss = gtk_check_button_new_with_label (
        _("Disable the screen saver"));
    gtk_box_pack_start (GTK_BOX (chooser), chooser->priv->killss,
                        FALSE, FALSE, MARGIN);
    gtk_widget_show (chooser->priv->killss);

    chooser->priv->hidecursor = gtk_check_button_new_with_label (
        _("Hide the mouse cursor"));
    gtk_box_pack_start (GTK_BOX (chooser), chooser->priv->hidecursor,
                        FALSE, FALSE, MARGIN);
    gtk_widget_show (chooser->priv->hidecursor);

    /* startup delay */
    hbox = gtk_box_new (GTK_ORIENTATION_HORIZONTAL, MARGIN);
    gtk_box_pack_start (GTK_BOX (chooser), hbox, FALSE, FALSE, MARGIN);
    gtk_widget_show (hbox);

    label = gtk_label_new (_("Delay before startup (seconds): "));
    gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, MARGIN);
    gtk_widget_show (label);

    chooser->priv->delay = gtk_spin_button_new_with_range (0, 60, 1);
    gtk_box_pack_start (GTK_BOX (hbox), chooser->priv->delay,
                        FALSE, FALSE, MARGIN);
    gtk_widget_show (chooser->priv->delay);

    hs = gtk_separator_new (GTK_ORIENTATION_HORIZONTAL);
    gtk_box_pack_start (GTK_BOX (chooser), hs, FALSE, TRUE, MARGIN);
    gtk_widget_show (hs);

    chooser->priv->head = LTOP_HEADING (ltop_heading_new ());
    gtk_box_pack_start (GTK_BOX (chooser), GTK_WIDGET (chooser->priv->head),
                        FALSE, TRUE, MARGIN);
    gtk_widget_show (GTK_WIDGET (chooser->priv->head));

    chooser->priv->config = gtk_label_new(_("Please select an item."));
    gtk_box_pack_start (GTK_BOX (chooser), chooser->priv->config,
                        TRUE, TRUE, MARGIN);
    gtk_widget_show (chooser->priv->config);

    g_signal_connect (G_OBJECT (chooser->priv->combo), "changed",
        G_CALLBACK (plugin_changed), (gpointer) chooser);
    g_signal_connect (G_OBJECT (chooser->priv->test_btn), "clicked",
        G_CALLBACK (test_btn_clicked), (gpointer) chooser);

    chooser->priv->modules = build_modules_list ();
    populate_store (chooser);
}

static void
as_chooser_finalize (GObject *object)
{
    ASChooser *chooser;

    g_return_if_fail (object != NULL);
    g_return_if_fail (AS_IS_CHOOSER (object));

    chooser = AS_CHOOSER (object);

    gtk_tree_model_foreach (GTK_TREE_MODEL (chooser->priv->store),
                            (GtkTreeModelForeachFunc) free_store,
                            NULL);
    g_object_unref (chooser->priv->store);
    g_list_free (chooser->priv->modules);
    g_free (chooser->priv->active_item);

    G_OBJECT_CLASS (as_chooser_parent_class)->finalize (object);
}

/* callbacks */
static void
plugin_changed (GtkWidget *combo, gpointer data)
{
    GetWidgetFunc  get_widget;
    FreeFunc       free_plugin;
    app           *item;
    ASChooser     *chooser = (ASChooser *) data;

    if (chooser->priv->active_item)
    {
        item = chooser->priv->active_item;
        g_module_symbol (item->module, "free_all",
                (gpointer *) &free_plugin);
        free_plugin();
    }

    item = get_active_plugin (chooser);
    chooser->priv->active_item = item;

    gtk_widget_destroy (chooser->priv->config);

    g_module_symbol (item->module, "get_config_widget",
            (gpointer *) &get_widget);
    chooser->priv->config = get_widget (chooser->priv->conffile);
    if (strcmp (item->icon, ""))
    {
        GdkPixbuf *pix = gdk_pixbuf_new_from_file_at_size (item->icon, 48, 48, NULL);
        ltop_heading_set_icon (chooser->priv->head, pix ? pix : NULL);
    } else {
        ltop_heading_set_icon_name (chooser->priv->head, NULL);
        ltop_heading_set_icon (chooser->priv->head, NULL);
    }
    ltop_heading_set_title (chooser->priv->head, item->display_name);
    gtk_box_pack_start (GTK_BOX (chooser), chooser->priv->config, FALSE, FALSE, MARGIN);
    gtk_widget_show (chooser->priv->config);
}

static void
test_btn_clicked (GtkWidget *w, gpointer data)
{
    app *item;
    GetCmdFunc get_cmd;
    gchar *tmp;
    ASChooser *chooser = (ASChooser *) data;

    item = get_active_plugin (chooser);
    if (strncmp (item->name, "none", 4) == 0)
        return;

    g_module_symbol (item->module, "get_cmd",
            (gpointer *) &get_cmd);
    tmp = g_strdup (get_cmd());
    if (tmp)
    {
        //xfce_exec (tmp, FALSE, TRUE, NULL);
        //FIXME: handle startup notification
        g_spawn_command_line_async (tmp, NULL);
        g_free (tmp);
    }
}

static GList *
build_modules_list ()
{
    GDir         *dirp;
    const gchar  *name;
    GList        *list = NULL;

    gchar  *fullpath = NULL;
    gchar  *sopath = NULL;
    gchar  *value = NULL;
    XfceRc *de;

    dirp = g_dir_open (DESKTOPDIR, 0, NULL);
    if (G_UNLIKELY(! dirp))
        return list;

    /* The None plugin first */
    app *newapp;
    newapp = g_new0 (app, 1);
    newapp->display_name = g_strdup (_("None"));
    newapp->name = g_strdup ("none");
    /* FIXME: should not be hardcoded! */
    newapp->icon = g_strdup ("/usr/share/icons/elementary-xfce/actions/32/gtk-cancel.svg");
    sopath = LIBDIR "/appstarter/plugins/none-plugin.so";
    newapp->module = g_module_open (sopath, G_MODULE_BIND_LOCAL);
    if (newapp->module)
        list = g_list_append (list, newapp);

    /* plugins from .desktop files */
    while ((name = g_dir_read_name (dirp)) != NULL)
    {
        fullpath = g_strconcat(DESKTOPDIR, "/", name, NULL);
        de = xfce_rc_simple_open (fullpath, TRUE);
        g_free (fullpath);

        if (G_UNLIKELY (!de))
            continue;
        newapp = g_new0 (app, 1);

        xfce_rc_set_group (de, "Desktop Entry");
        newapp->display_name = g_strdup (
                xfce_rc_read_entry (de, "GenericName", ""));
        newapp->name = g_strdup (xfce_rc_read_entry (de, "Name", ""));
        newapp->icon = g_strdup (xfce_rc_read_entry (de, "Icon", ""));
        sopath = g_strdup (xfce_rc_read_entry (de, "Exec", ""));

        xfce_rc_close (de);

        if (!sopath)
            continue;

        newapp->module = g_module_open (sopath, G_MODULE_BIND_LOCAL);
        g_free (sopath);
        if (G_LIKELY (newapp->module))
            list = g_list_append (list, newapp);
        else
        {
            g_printf("Module import error: %s\n", g_module_error());
            g_free (newapp);
        }
        g_free (value);
    }
    g_dir_close(dirp);

    return list;
}

static void
append_to_store (gpointer data, gpointer user_data)
{
    static int compt = 0;
    GtkTreeIter iter;
    GdkPixbuf *pix;
    ASChooser *chooser = (ASChooser *) user_data;
    app *item = (app *) data;

    pix = gdk_pixbuf_new_from_file_at_size (item->icon, 24, 24, NULL);
    gtk_list_store_append (chooser->priv->store, &iter);
    gtk_list_store_set (chooser->priv->store, &iter,
                        STORE_COMPT, compt++,
                        STORE_ICON, pix,
                        STORE_NAME, item->display_name,
                        STORE_ITEM, (gpointer *)item,
                        -1);
}

static void
populate_store (ASChooser *chooser)
{
    g_list_foreach (chooser->priv->modules, append_to_store, (gpointer) chooser);
}

static app *
get_active_plugin (ASChooser *chooser)
{
    GtkTreeModel  *model;
    GtkTreeIter    iter;
    app *item;

    model = gtk_combo_box_get_model(GTK_COMBO_BOX (chooser->priv->combo));
    gtk_combo_box_get_active_iter (GTK_COMBO_BOX (chooser->priv->combo), &iter);
    gtk_tree_model_get (model, &iter, STORE_ITEM, &item, -1);

    return item;
}

static void
set_default (gpointer data, gpointer user_data)
{
    static int compt = 0;
    app *item = (app *) data;
    ASChooser *chooser = (ASChooser *) user_data;

    if (!strcmp (item->name, chooser->priv->selected))
        gtk_combo_box_set_active (GTK_COMBO_BOX (chooser->priv->combo), compt);

    compt++;
}

static void
read_config (ASChooser *chooser)
{
    XfceRc *rc = as_rc_open (chooser->priv->conffile, TRUE);
    chooser->priv->selected = NULL;

    if (!rc)
    {
        chooser->priv->selected = "none";
        return;
    }

    xfce_rc_set_group (rc, "default");
    chooser->priv->selected = xfce_rc_read_entry (rc, "selected", "none");
//    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (chooser->priv->onlyme),
//        xfce_rc_read_bool_entry (rc, "nodesktop", FALSE));
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (chooser->priv->killss),
        xfce_rc_read_bool_entry (rc, "noscreensaver", TRUE));
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (chooser->priv->hidecursor),
        xfce_rc_read_bool_entry (rc, "hidecursor", FALSE));
    gtk_spin_button_set_value (GTK_SPIN_BUTTON (chooser->priv->delay),
        xfce_rc_read_int_entry (rc, "delay", 0));

    xfce_rc_close (rc);

    /* set the default plugin */
    g_list_foreach (chooser->priv->modules, set_default, (gpointer) chooser);
}

static gboolean
free_store (GtkTreeModel *model, GtkTreePath *path,
            GtkTreeIter *iter,gpointer data)
{
    app *item;
    gtk_tree_model_get (model, iter, STORE_ITEM, &item, -1);
    g_free (item->name);
    g_free (item->icon);
    g_free (item->display_name);
    g_module_close (item->module);
    g_free (item);

    return FALSE;
}

/**
 * SECTION:as-chooser
 * @short_description: the applicaction chooser widget
 * @include libappstarter/as.h
 *
 */

/**
 * as_chooser_new:
 * @file: the configuration file to use
 *
 * Creates a new ASChooser #GtkWidget.
 *
 * Returns: a GtkWidget
 */
GtkWidget *
as_chooser_new (const gchar *file)
{
    g_return_val_if_fail (file != NULL, NULL);

    ASChooser *chooser;
    chooser = g_object_new (as_chooser_get_type (), NULL);

    chooser->priv->conffile = g_strdup (file);
    read_config (chooser);

    return GTK_WIDGET (chooser);
}

/**
 * as_chooser_save
 * @chooser: an ASChooser
 *
 * Save the configuration to file.
 *
 * Returns: TRUE on succes, FALSE otherwise
 */
gboolean
as_chooser_save (ASChooser *chooser)
{
    XfceRc *rc;
    GetCmdFunc get_cmd;
    app *item;

    rc = as_rc_open (chooser->priv->conffile, FALSE);
    if (G_UNLIKELY (!rc))
        return FALSE;

    xfce_rc_set_group (rc, "default");
    item = get_active_plugin (chooser);
    xfce_rc_write_entry (rc, "selected", item->name);
//    xfce_rc_write_bool_entry (rc, "nodesktop",
//        gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (
//            chooser->priv->onlyme)));
    xfce_rc_write_bool_entry (rc, "noscreensaver",
        gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (
            chooser->priv->killss)));
    xfce_rc_write_bool_entry (rc, "hidecursor",
        gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (
            chooser->priv->hidecursor)));
    xfce_rc_write_int_entry (rc, "delay",
        gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON (
            chooser->priv->delay)));

    g_module_symbol (item->module, "get_cmd", (gpointer *) &get_cmd);
    xfce_rc_write_entry (rc, "command", get_cmd ());
    xfce_rc_close (rc);

    return TRUE;
}
